@@ -328,8 +328,7 @@ gboolean
 udisks_linux_drive_ata_update (UDisksLinuxDriveAta    *drive,
                                UDisksLinuxDriveObject *object)
 {
-  UDisksLinuxDevice *device
-;
+  UDisksLinuxDevice *device;
   device = udisks_linux_drive_object_get_device (object, TRUE /* get_hw */);
   if (device == NULL)
     goto out;
@@ -486,13 +485,19 @@ static gboolean update_io_stats (UDisksLinuxDriveAta *drive, UDisksLinuxDevice *
     }
   else
     {
-      fscanf (statf, "%lu %*u %*u %*u %lu", &drive_read, &drive_write);
+      if (fscanf (statf, "%lu %*u %*u %*u %lu", &drive_read, &drive_write) != 2)
+        {
+          udisks_warning ("Failed to read %s\n", statpath);
+        }
+      else
+        {
+          noio = drive_read == drive->drive_read && drive_write == drive->drive_write;
+          udisks_debug ("drive_read=%lu, drive_write=%lu, old_drive_read=%lu, old_drive_write=%lu\n",
+                        drive_read, drive_write, drive->drive_read, drive->drive_write);
+          drive->drive_read = drive_read;
+          drive->drive_write = drive_write;
+        }
       fclose (statf);
-      noio = drive_read == drive->drive_read && drive_write == drive->drive_write;
-      udisks_debug ("drive_read=%lu, drive_write=%lu, old_drive_read=%lu, old_drive_write=%lu\n",
-                    drive_read, drive_write, drive->drive_read, drive->drive_write);
-      drive->drive_read = drive_read;
-      drive->drive_write = drive_write;
     }
   return noio;
 }
@@ -701,10 +706,10 @@ udisks_linux_drive_ata_refresh_smart_sync (UDisksLinuxDriveAta  *drive,
  * Returns: %TRUE if the operation succeed, %FALSE if @error is set.
  */
 gboolean
-udisks_linux_drive_ata_smart_selftest_sync (UDisksLinuxDriveAta     *drive,
-                                            const gchar             *type,
-                                            GCancellable            *cancellable,
-                                            GError                 **error)
+udisks_linux_drive_ata_smart_selftest_sync (UDisksLinuxDriveAta  *drive,
+                                            const gchar          *type,
+                                            GCancellable         *cancellable,
+                                            GError              **error)
 {
   UDisksLinuxDriveObject  *object;
   UDisksLinuxDevice *device = NULL;
@@ -862,9 +867,9 @@ handle_smart_update (UDisksDriveAta        *_drive,
                                                   NULL, /* cancellable */
                                                   &error))
     {
-      udisks_warning ("Error updating ATA smart for %s: %s (%s, %d)",
-                      g_dbus_object_get_object_path (G_DBUS_OBJECT (object)),
-                      error->message, g_quark_to_string (error->domain), error->code);
+      udisks_debug ("Error updating ATA smart for %s: %s (%s, %d)",
+                    g_dbus_object_get_object_path (G_DBUS_OBJECT (object)),
+                    error->message, g_quark_to_string (error->domain), error->code);
       g_dbus_method_invocation_take_error (invocation, error);
       goto out;
     }
@@ -1180,7 +1185,7 @@ handle_smart_selftest_start (UDisksDriveAta        *_drive,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -1233,6 +1238,7 @@ handle_smart_selftest_start (UDisksDriveAta        *_drive,
                                                                                     g_object_ref (drive),
                                                                                     g_object_unref,
                                                                                     NULL)); /* GCancellable */
+      udisks_threaded_job_start (drive->selftest_job);
     }
   G_UNLOCK (object_lock);
 
@@ -1328,9 +1334,10 @@ handle_pm_get_state (UDisksDriveAta        *_drive,
 /* ---------------------------------------------------------------------------------------------------- */
 
 static gboolean
-handle_pm_standby (UDisksDriveAta        *_drive,
-                   GDBusMethodInvocation *invocation,
-                   GVariant              *options)
+handle_pm_standby_wakeup (UDisksDriveAta        *_drive,
+                          GDBusMethodInvocation *invocation,
+                          GVariant              *options,
+                          gboolean              do_wakeup)
 {
   UDisksLinuxDriveAta *drive = UDISKS_LINUX_DRIVE_ATA (_drive);
   UDisksLinuxDriveObject *object = NULL;
@@ -1343,6 +1350,8 @@ handle_pm_standby (UDisksDriveAta        *_drive,
   const gchar *message;
   const gchar *action_id;
   uid_t caller_uid;
+  guchar buf[4096];
+  int open_flags = (do_wakeup) ? (O_RDONLY) : (O_RDONLY|O_NONBLOCK);
 
   object = udisks_daemon_util_dup_object (drive, &error);
   if (object == NULL)
@@ -1384,17 +1393,19 @@ handle_pm_standby (UDisksDriveAta        *_drive,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
   /* Translators: Shown in authentication dialog when the user
-   * tries to put a drive into standby mode.
+   * tries to wake up a drive from standby mode or tries to put a drive into
+   * standby mode.
    *
    * Do not translate $(drive), it's a placeholder and
    * will be replaced by the name of the drive/device in question
    */
-  message = N_("Authentication is required to put $(drive) in standby mode");
+  message = (do_wakeup) ? N_("Authentication is required to wake up $(drive) from standby mode") :
+                          N_("Authentication is required to put $(drive) in standby mode");
   action_id = "org.freedesktop.udisks2.ata-standby";
   if (udisks_block_get_hint_system (block))
     {
@@ -1422,7 +1433,7 @@ handle_pm_standby (UDisksDriveAta        *_drive,
       goto out;
     }
 
-  fd = open (g_udev_device_get_device_file (device->udev_device), O_RDONLY|O_NONBLOCK);
+  fd = open (g_udev_device_get_device_file (device->udev_device), open_flags);
   if (fd == -1)
     {
       g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
@@ -1431,23 +1442,36 @@ handle_pm_standby (UDisksDriveAta        *_drive,
       goto out;
     }
 
-  {
-    /* ATA8: 7.55 STANDBY IMMEDIATE - E0h, Non-Data */
-    UDisksAtaCommandInput input = {.command = 0xe0};
-    UDisksAtaCommandOutput output = {0};
-    if (!udisks_ata_send_command_sync (fd,
-                                       -1,
-                                       UDISKS_ATA_COMMAND_PROTOCOL_NONE,
-                                       &input,
-                                       &output,
-                                       &error))
+  if (do_wakeup)
+   {
+      if (read (fd, buf, sizeof (buf)) != sizeof (buf))
+        {
+          g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
+                                                 "Error reading %d bytes from %s: %m",
+                                                 (gint) sizeof (buf),
+                                                 g_udev_device_get_device_file (device->udev_device));
+          goto out;
+        }
+      udisks_drive_ata_complete_pm_wakeup (_drive, invocation);
+   }
+  else
+   {
+     /* ATA8: 7.55 STANDBY IMMEDIATE - E0h, Non-Data */
+     UDisksAtaCommandInput input = {.command = 0xe0};
+     UDisksAtaCommandOutput output = {0};
+     if (!udisks_ata_send_command_sync (fd,
+                                        -1,
+                                        UDISKS_ATA_COMMAND_PROTOCOL_NONE,
+                                        &input,
+                                        &output,
+                                        &error))
       {
         g_prefix_error (&error, "Error sending ATA command STANDBY IMMEDIATE: ");
         g_dbus_method_invocation_take_error (invocation, error);
         goto out;
       }
-    udisks_drive_ata_complete_pm_standby (_drive, invocation);
-  }
+     udisks_drive_ata_complete_pm_standby (_drive, invocation);
+   }
 
  out:
   if (fd != -1)
@@ -1458,6 +1482,15 @@ handle_pm_standby (UDisksDriveAta        *_drive,
   return TRUE; /* returning TRUE means that we handled the method invocation */
 }
 
+
+static gboolean
+handle_pm_standby (UDisksDriveAta        *_drive,
+                   GDBusMethodInvocation *invocation,
+                   GVariant              *options)
+{
+ return handle_pm_standby_wakeup (_drive, invocation, options, FALSE);
+}
+
 /* ---------------------------------------------------------------------------------------------------- */
 
 static gboolean
@@ -1465,124 +1498,7 @@ handle_pm_wakeup (UDisksDriveAta        *_drive,
                   GDBusMethodInvocation *invocation,
                   GVariant              *options)
 {
-  UDisksLinuxDriveAta *drive = UDISKS_LINUX_DRIVE_ATA (_drive);
-  UDisksLinuxDriveObject *object = NULL;
-  UDisksLinuxBlockObject *block_object = NULL;
-  UDisksBlock *block = NULL;
-  UDisksDaemon *daemon;
-  UDisksLinuxDevice *device = NULL;
-  gint fd = -1;
-  GError *error = NULL;
-  const gchar *message;
-  const gchar *action_id;
-  guchar buf[4096];
-  uid_t caller_uid;
-
-  object = udisks_daemon_util_dup_object (drive, &error);
-  if (object == NULL)
-    {
-      g_dbus_method_invocation_take_error (invocation, error);
-      goto out;
-    }
-
-  block_object = udisks_linux_drive_object_get_block (object, FALSE);
-  if (block_object == NULL)
-    {
-      g_dbus_method_invocation_return_error (invocation,
-                                             UDISKS_ERROR,
-                                             UDISKS_ERROR_FAILED,
-                                             "Unable to find block device for drive");
-      goto out;
-    }
-  block = udisks_object_peek_block (UDISKS_OBJECT (block_object));
-
-  daemon = udisks_linux_drive_object_get_daemon (object);
-
-  if (!udisks_drive_ata_get_pm_supported (UDISKS_DRIVE_ATA (drive)) ||
-      !udisks_drive_ata_get_pm_enabled (UDISKS_DRIVE_ATA (drive)))
-    {
-      g_dbus_method_invocation_return_error (invocation,
-                                             UDISKS_ERROR,
-                                             UDISKS_ERROR_FAILED,
-                                             "PM is not supported or enabled");
-      goto out;
-    }
-
-  error = NULL;
-  if (!udisks_daemon_util_get_caller_uid_sync (daemon,
-                                               invocation,
-                                               NULL /* GCancellable */,
-                                               &caller_uid,
-                                               NULL,
-                                               NULL,
-                                               &error))
-    {
-      g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
-      goto out;
-    }
-
-  /* Translators: Shown in authentication dialog when the user
-   * tries to wake up a drive from standby mode.
-   *
-   * Do not translate $(drive), it's a placeholder and
-   * will be replaced by the name of the drive/device in question
-   */
-  message = N_("Authentication is required to wake up $(drive) from standby mode");
-  action_id = "org.freedesktop.udisks2.ata-standby";
-  if (udisks_block_get_hint_system (block))
-    {
-      action_id = "org.freedesktop.udisks2.ata-standby-system";
-    }
-  else if (!udisks_daemon_util_on_user_seat (daemon, UDISKS_OBJECT (object), caller_uid))
-    {
-      action_id = "org.freedesktop.udisks2.ata-standby-other-seat";
-    }
-
-  /* Check that the user is authorized */
-  if (!udisks_daemon_util_check_authorization_sync (daemon,
-                                                    UDISKS_OBJECT (object),
-                                                    action_id,
-                                                    options,
-                                                    message,
-                                                    invocation))
-    goto out;
-
-  device = udisks_linux_drive_object_get_device (object, TRUE /* get_hw */);
-  if (device == NULL)
-    {
-      g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
-                                             "No udev device");
-      goto out;
-    }
-
-  fd = open (g_udev_device_get_device_file (device->udev_device), O_RDONLY);
-  if (fd == -1)
-    {
-      g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
-                                             "Error opening device file %s: %m",
-                                             g_udev_device_get_device_file (device->udev_device));
-      goto out;
-    }
-
-  if (read (fd, buf, sizeof (buf)) != sizeof (buf))
-    {
-      g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
-                                             "Error reading %d bytes from %s: %m",
-                                             (gint) sizeof (buf),
-                                             g_udev_device_get_device_file (device->udev_device));
-      goto out;
-    }
-
-  udisks_drive_ata_complete_pm_wakeup (_drive, invocation);
-
- out:
-  if (fd != -1)
-    close (fd);
-  g_clear_object (&device);
-  g_clear_object (&block_object);
-  g_clear_object (&object);
-  return TRUE; /* returning TRUE means that we handled the method invocation */
+  return handle_pm_standby_wakeup (_drive, invocation, options, TRUE);
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -1678,7 +1594,7 @@ apply_configuration_thread_func (gpointer user_data)
   fd = open (device_file, O_RDWR|O_NONBLOCK);
   if (fd == -1)
     {
-      udisks_error ("Error opening device file %s: %m", device_file);
+      udisks_critical ("Error opening device file %s: %m", device_file);
       goto out;
     }
 
@@ -1694,7 +1610,7 @@ apply_configuration_thread_func (gpointer user_data)
                                          &output,
                                          &error))
         {
-          udisks_error ("Error sending ATA command IDLE (timeout=%d) to %s: %s (%s, %d)",
+          udisks_critical ("Error sending ATA command IDLE (timeout=%d) to %s: %s (%s, %d)",
                         data->ata_pm_standby, device_file,
                         error->message, g_quark_to_string (error->domain), error->code);
           g_clear_error (&error);
@@ -1728,7 +1644,7 @@ apply_configuration_thread_func (gpointer user_data)
                                          &output,
                                          &error))
         {
-          udisks_error ("Error sending ATA command SET FEATURES, sub-command 0x%02x (ata_apm_level=%d) to %s: %s (%s, %d)",
+          udisks_critical ("Error sending ATA command SET FEATURES, sub-command 0x%02x (ata_apm_level=%d) to %s: %s (%s, %d)",
                         (guint) input.feature, data->ata_apm_level, device_file,
                         error->message, g_quark_to_string (error->domain), error->code);
           g_clear_error (&error);
@@ -1747,7 +1663,7 @@ apply_configuration_thread_func (gpointer user_data)
        */
       UDisksAtaCommandInput input = {.command = 0xef, .feature = 0x42, .count = data->ata_aam_level};
       UDisksAtaCommandOutput output = {0};
-      if (data->ata_apm_level == 0xff)
+      if (data->ata_aam_level == 0xff)
         {
           input.feature = 0xc2;
           input.count = 0x00;
@@ -1759,7 +1675,7 @@ apply_configuration_thread_func (gpointer user_data)
                                          &output,
                                          &error))
         {
-          udisks_error ("Error sending ATA command SET FEATURES, sub-command 0x%02x (ata_aam_level=%d) to %s: %s (%s, %d)",
+          udisks_critical ("Error sending ATA command SET FEATURES, sub-command 0x%02x (ata_aam_level=%d) to %s: %s (%s, %d)",
                         (guint) input.feature, data->ata_aam_level, device_file,
                         error->message, g_quark_to_string (error->domain), error->code);
           g_clear_error (&error);
@@ -1787,7 +1703,7 @@ apply_configuration_thread_func (gpointer user_data)
                                          &output,
                                          &error))
         {
-          udisks_error ("Error sending ATA command SET FEATURES, sub-command 0x%02x to %s: %s (%s, %d)",
+          udisks_critical ("Error sending ATA command SET FEATURES, sub-command 0x%02x to %s: %s (%s, %d)",
                         (guint) input.feature, device_file,
                         error->message, g_quark_to_string (error->domain), error->code);
           g_clear_error (&error);
@@ -1816,7 +1732,7 @@ apply_configuration_thread_func (gpointer user_data)
                                          &output,
                                          &error))
         {
-          udisks_error ("Error sending ATA command SET FEATURES, sub-command 0x%02x to %s: %s (%s, %d)",
+          udisks_critical ("Error sending ATA command SET FEATURES, sub-command 0x%02x to %s: %s (%s, %d)",
                         (guint) input.feature, device_file,
                         error->message, g_quark_to_string (error->domain), error->code);
           g_clear_error (&error);
@@ -1846,9 +1762,9 @@ apply_configuration_thread_func (gpointer user_data)
  * wait for the thread to terminate.
  */
 void
-udisks_linux_drive_ata_apply_configuration (UDisksLinuxDriveAta     *drive,
-                                            UDisksLinuxDevice       *device,
-                                            GVariant                *configuration)
+udisks_linux_drive_ata_apply_configuration (UDisksLinuxDriveAta *drive,
+                                            UDisksLinuxDevice   *device,
+                                            GVariant            *configuration)
 {
   gboolean has_conf = FALSE;
   ApplyConfData *data = NULL;
@@ -1951,10 +1867,10 @@ on_secure_erase_update_progress_timeout (gpointer user_data)
  * Returns: %TRUE if the operation succeeded, %FALSE if @error is set.
  */
 gboolean
-udisks_linux_drive_ata_secure_erase_sync (UDisksLinuxDriveAta     *drive,
-                                          uid_t                    caller_uid,
-                                          gboolean                 enhanced,
-                                          GError                 **error)
+udisks_linux_drive_ata_secure_erase_sync (UDisksLinuxDriveAta  *drive,
+                                          uid_t                 caller_uid,
+                                          gboolean              enhanced,
+                                          GError              **error)
 {
   gboolean ret = FALSE;
   UDisksDrive *_drive = NULL;
@@ -2210,7 +2126,7 @@ udisks_linux_drive_ata_secure_erase_sync (UDisksLinuxDriveAta     *drive,
                                          &output,
                                          &cleanup_error))
         {
-          udisks_error ("Failed to clear user password '%s' on %s (%s) while attemping clean-up after a failed secure erase operation. You may need to manually unlock the drive. The error was: %s (%s, %d)",
+          udisks_critical ("Failed to clear user password '%s' on %s (%s) while attemping clean-up after a failed secure erase operation. You may need to manually unlock the drive. The error was: %s (%s, %d)",
                         pass,
                         device_file,
                         udisks_drive_get_id (_drive),
@@ -2316,7 +2232,7 @@ handle_security_erase_unit (UDisksDriveAta        *_drive,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -2345,7 +2261,7 @@ handle_security_erase_unit (UDisksDriveAta        *_drive,
   if (!udisks_linux_drive_ata_secure_erase_sync (drive, caller_uid, enhanced, &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
@@ -2409,7 +2325,7 @@ handle_smart_set_enabled (UDisksDriveAta        *_drive,
                                                &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
-      g_error_free (error);
+      g_clear_error (&error);
       goto out;
     }
 
