@@ -26,7 +26,6 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
-#include <sys/ioctl.h>
 
 #include <pwd.h>
 #include <grp.h>
@@ -34,7 +33,9 @@
 #include <string.h>
 #include <stdlib.h>
 
-#include <linux/loop.h>
+#include <blockdev/loop.h>
+#include <blockdev/fs.h>
+#include <blockdev/mdraid.h>
 
 #include "udiskslogging.h"
 #include "udiskslinuxmanager.h"
@@ -45,6 +46,7 @@
 #include "udiskslinuxdevice.h"
 #include "udisksmodulemanager.h"
 #include "udiskslinuxfsinfo.h"
+#include "udiskssimplejob.h"
 
 /**
  * SECTION:udiskslinuxmanager
@@ -307,11 +309,8 @@ handle_loop_setup (UDisksManager          *object,
   gchar proc_path[64];
   gchar path[8192];
   ssize_t path_len;
-  gint loop_fd = -1;
-  gint loop_control_fd = -1;
-  gint allocated_loop_number = -1;
   gchar *loop_device = NULL;
-  struct loop_info64 li64;
+  const gchar *loop_name = NULL;
   UDisksObject *loop_object = NULL;
   gboolean option_read_only = FALSE;
   gboolean option_no_part_scan = FALSE;
@@ -386,69 +385,30 @@ handle_loop_setup (UDisksManager          *object,
   if (fstat (fd, &fd_statbuf) == 0)
     fd_statbuf_valid = TRUE;
 
-  /* serialize access to /dev/loop-control */
-  g_mutex_lock (&(manager->lock));
-
-  loop_control_fd = open ("/dev/loop-control", O_RDWR);
-  if (loop_control_fd == -1)
+  error = NULL;
+  if (!bd_loop_setup_from_fd (fd,
+                              option_offset,
+                              option_size,
+                              option_read_only,
+                              !option_no_part_scan,
+                              &loop_name,
+                              &error))
     {
-      g_dbus_method_invocation_return_error (invocation,
-                                             UDISKS_ERROR,
-                                             UDISKS_ERROR_FAILED,
-                                             "Error opening /dev/loop-control: %m");
-      g_mutex_unlock (&(manager->lock));
+      g_prefix_error (&error, "Error creating loop device: ");
+      g_dbus_method_invocation_take_error (invocation, error);
       goto out;
     }
 
-  allocated_loop_number = ioctl (loop_control_fd, LOOP_CTL_GET_FREE);
-  if (allocated_loop_number < 0)
-    {
-      g_dbus_method_invocation_return_error (invocation,
-                                             UDISKS_ERROR,
-                                             UDISKS_ERROR_FAILED,
-                                             "Error allocating free loop device: %m");
-      g_mutex_unlock (&(manager->lock));
-      goto out;
-    }
+  loop_device = g_strdup_printf ("/dev/%s", loop_name);
 
-  loop_device = g_strdup_printf ("/dev/loop%d", allocated_loop_number);
-  loop_fd = open (loop_device, option_read_only ? O_RDONLY : O_RDWR);
-  if (loop_fd == -1)
-    {
-      g_dbus_method_invocation_return_error (invocation,
-                                             UDISKS_ERROR,
-                                             UDISKS_ERROR_FAILED,
-                                             "Cannot open %s: %m", loop_device);
-      g_mutex_unlock (&(manager->lock));
-      goto out;
-    }
-
-  /* update the loop file - need to do this before getting the uevent for the device  */
+  /* Update the udisks loop state file (/run/udisks2/loop) with information
+   * about the new loop device created by us.
+   */
   udisks_state_add_loop (udisks_daemon_get_state (manager->daemon),
-                         loop_device,
-                         path,
-                         fd_statbuf_valid ? fd_statbuf.st_dev : 0,
-                         caller_uid);
-
-  memset (&li64, '\0', sizeof (li64));
-  strncpy ((char *) li64.lo_file_name, path, LO_NAME_SIZE - 1);
-  if (option_read_only)
-    li64.lo_flags |= LO_FLAGS_READ_ONLY;
-  if (!option_no_part_scan)
-    li64.lo_flags |= 8; /* Use LO_FLAGS_PARTSCAN when 3.2 has been out for a while */
-  li64.lo_offset = option_offset;
-  li64.lo_sizelimit = option_size;
-  if (ioctl (loop_fd, LOOP_SET_FD, fd) < 0 || ioctl (loop_fd, LOOP_SET_STATUS64, &li64) < 0)
-    {
-      g_dbus_method_invocation_return_error (invocation,
-                                             UDISKS_ERROR,
-                                             UDISKS_ERROR_FAILED,
-                                             "Error setting up loop device %s: %m",
-                                             loop_device);
-      g_mutex_unlock (&(manager->lock));
-      goto out;
-    }
-  g_mutex_unlock (&(manager->lock));
+                        loop_device,
+                        path,
+                        fd_statbuf_valid ? fd_statbuf.st_dev : 0,
+                        caller_uid);
 
   /* Determine the resulting object */
   error = NULL;
@@ -482,10 +442,7 @@ handle_loop_setup (UDisksManager          *object,
   if (loop_object != NULL)
     g_object_unref (loop_object);
   g_free (loop_device);
-  if (loop_control_fd != -1)
-    close (loop_control_fd);
-  if (loop_fd != -1)
-    close (loop_fd);
+  g_free ((gpointer) loop_name);
   if (fd != -1)
     close (fd);
   return TRUE; /* returning TRUE means that we handled the method invocation */
@@ -530,27 +487,6 @@ wait_for_array_object (UDisksDaemon *daemon,
   return ret;
 }
 
-static gchar* md_node_from_name (const gchar *name, GError **error) {
-    gchar *symlink = NULL;
-    gchar *ret = NULL;
-    gchar *md_path = g_strdup_printf ("/dev/md/%s", name);
-
-    symlink = g_file_read_link (md_path, error);
-    if (!symlink) {
-        /* error is already populated */
-        g_free (md_path);
-        return NULL;
-    }
-
-    g_strstrip (symlink);
-    ret = g_path_get_basename (symlink);
-
-    g_free (symlink);
-    g_free (md_path);
-
-    return ret;
-}
-
 static const gchar *raid_level_whitelist[] = {"raid0", "raid1", "raid4", "raid5", "raid6", "raid10", NULL};
 
 static gboolean
@@ -573,19 +509,25 @@ handle_mdraid_create (UDisksManager         *_object,
   GList *l;
   guint n;
   gchar *array_name = NULL;
-  GString *str = NULL;
-  gint status;
-  gchar *error_message = NULL;
   gchar *raid_device_file = NULL;
-  gchar *raid_device_name = NULL;
+  gchar *raid_node = NULL;
   struct stat statbuf;
   dev_t raid_device_num;
+  UDisksBaseJob *job = NULL;
+  const gchar **disks = NULL;
+  guint disks_top = 0;
+  gboolean success = FALSE;
 
-  error = NULL;
-  if (!udisks_daemon_util_get_caller_uid_sync (manager->daemon, invocation, NULL /* GCancellable */, &caller_uid, NULL, NULL, &error))
+  if (!udisks_daemon_util_get_caller_uid_sync (manager->daemon,
+                                               invocation,
+                                               NULL /* GCancellable */,
+                                               &caller_uid,
+                                               NULL, NULL,
+                                               &error))
     {
       g_dbus_method_invocation_return_gerror (invocation, error);
       g_clear_error (&error);
+      success = FALSE;
       goto out;
     }
 
@@ -601,7 +543,25 @@ handle_mdraid_create (UDisksManager         *_object,
                                                     arg_options,
                                                     message,
                                                     invocation))
-    goto out;
+    {
+      success = FALSE;
+      goto out;
+    }
+
+  /* Authentication checked -- lets create the job */
+  job = udisks_daemon_launch_simple_job (manager->daemon,
+                                         NULL,
+                                         "mdraid-create",
+                                         caller_uid,
+                                         NULL);
+
+  if (job == NULL)
+    {
+      g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
+                                             "Failed to create a job object");
+      success = FALSE;
+      goto out;
+    }
 
   /* validate level */
   for (n = 0; raid_level_whitelist[n] != NULL; n++)
@@ -613,6 +573,7 @@ handle_mdraid_create (UDisksManager         *_object,
     {
       g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
                                              "Unsupported RAID level %s", arg_level);
+      success = FALSE;
       goto out;
     }
 
@@ -621,14 +582,16 @@ handle_mdraid_create (UDisksManager         *_object,
     {
       g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
                                              "Chunk %" G_GUINT64_FORMAT " is not a multiple of 4KiB", arg_chunk);
+      success = FALSE;
       goto out;
     }
 
-  /* validate name */
+  /* validate chunk for raid1 */
   if (g_strcmp0 (arg_level, "raid1") == 0 && arg_chunk != 0)
     {
       g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
                                              "Chunk must be zero for level 'raid1'");
+      success = FALSE;
       goto out;
     }
 
@@ -636,7 +599,8 @@ handle_mdraid_create (UDisksManager         *_object,
   if (strlen (arg_name) > 32)
     {
       g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
-                                             "Name is invalid");
+                                             "Name cannot be longer than 32 characters");
+      success = FALSE;
       goto out;
     }
 
@@ -647,6 +611,7 @@ handle_mdraid_create (UDisksManager         *_object,
     {
       g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
                                              "Must have at least two devices");
+      success = FALSE;
       goto out;
     }
 
@@ -671,6 +636,7 @@ handle_mdraid_create (UDisksManager         *_object,
                                                  UDISKS_ERROR_FAILED,
                                                  "Invalid object path %s at index %u",
                                                  arg_blocks[n], n);
+          success = FALSE;
           goto out;
         }
 
@@ -682,6 +648,7 @@ handle_mdraid_create (UDisksManager         *_object,
                                                  UDISKS_ERROR_FAILED,
                                                  "Object path %s for index %u is not a block device",
                                                  arg_blocks[n], n);
+          success = FALSE;
           goto out;
         }
 
@@ -695,6 +662,7 @@ handle_mdraid_create (UDisksManager         *_object,
                                                  "Error opening device %s: %m",
                                                  device_file);
           g_free (device_file);
+          success = FALSE;
           goto out;
         }
       close (fd);
@@ -709,52 +677,26 @@ handle_mdraid_create (UDisksManager         *_object,
   for (l = blocks; l != NULL; l = l->next)
     {
       UDisksBlock *block = UDISKS_BLOCK (l->data);
-      UDisksObject *object_for_block;
-      gchar *escaped_device;
-      object_for_block = udisks_daemon_util_dup_object (block, &error);
-      if (object_for_block == NULL)
+      if (!bd_fs_wipe (udisks_block_get_device (block), TRUE, &error))
         {
-          g_dbus_method_invocation_return_gerror (invocation, error);
-          g_clear_error (&error);
-          goto out;
+          /* no signature to remove, ignore */
+          if (g_error_matches (error, BD_FS_ERROR, BD_FS_ERROR_NOFS))
+            g_clear_error (&error);
+          else
+            {
+              g_prefix_error (&error,
+                              "Error wiping device %s to be used in the RAID array:",
+                              udisks_block_get_device (block));
+              g_dbus_method_invocation_take_error (invocation, error);
+              success = FALSE;
+              goto out;
+            }
         }
-      escaped_device = udisks_daemon_util_escape (udisks_block_get_device (block));
-      if (!udisks_daemon_launch_spawned_job_sync (manager->daemon,
-                                                  object_for_block,
-                                                  "format-erase", caller_uid,
-                                                  NULL, /* cancellable */
-                                                  0,    /* uid_t run_as_uid */
-                                                  0,    /* uid_t run_as_euid */
-                                                  &status,
-                                                  &error_message,
-                                                  NULL, /* input_string */
-                                                  "wipefs -a \"%s\"",
-                                                  escaped_device))
-        {
-          g_dbus_method_invocation_return_error (invocation,
-                                                 UDISKS_ERROR,
-                                                 UDISKS_ERROR_FAILED,
-                                                 "Error wiping device %s to be used in a RAID array: %s",
-                                                 udisks_block_get_device (block),
-                                                 error_message);
-          g_free (error_message);
-          g_object_unref (object_for_block);
-          g_free (escaped_device);
-          goto out;
-        }
-      g_object_unref (object_for_block);
-      g_free (escaped_device);
     }
 
-  /* Create the array... */
-  str = g_string_new ("mdadm");
-
   /* we have name from the user */
   if (strlen (arg_name) > 0)
-    {
-      array_name = udisks_daemon_util_escape (arg_name);
-      g_string_append_printf (str, " --create \"%s\"", array_name);
-    }
+      array_name = g_strdup (arg_name);
   /* we don't have name, get next 'free' /dev/mdX device */
   else
     {
@@ -763,60 +705,41 @@ handle_mdraid_create (UDisksManager         *_object,
         {
           g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
                                                  "Unable to find free MD device");
+          success = FALSE;
           goto out;
         }
-      g_string_append_printf (str, " --create %s", array_name);
     }
 
-  g_string_append_printf (str, " --run");
-  if (arg_chunk > 0)
-    g_string_append_printf (str, " --chunk %" G_GUINT64_FORMAT, (guint64) (arg_chunk / 1024LL));
-  g_string_append_printf (str, " --level %s", arg_level);
-  g_string_append_printf (str, " --raid-devices %u", num_devices);
+  /* names of members as gchar** for libblockdev */
+  disks = g_new0 (const gchar*, g_list_length (blocks) + 1);
   for (l = blocks; l != NULL; l = l->next)
     {
       UDisksBlock *block = UDISKS_BLOCK (l->data);
-      gchar *escaped_device;
-      escaped_device = udisks_daemon_util_escape (udisks_block_get_device (block));
-      g_string_append_printf (str, " \"%s\"", escaped_device);
-      g_free (escaped_device);
+      disks[disks_top++] = udisks_block_dup_device (block);
     }
+  disks[disks_top] = NULL;
 
-  if (!udisks_daemon_launch_spawned_job_sync (manager->daemon,
-                                              NULL,
-                                              "mdraid-create", caller_uid,
-                                              NULL, /* cancellable */
-                                              0,    /* uid_t run_as_uid */
-                                              0,    /* uid_t run_as_euid */
-                                              &status,
-                                              &error_message,
-                                              NULL, /* input_string */
-                                              "%s",
-                                              str->str))
+  if (!bd_md_create (array_name, arg_level, disks, 0, NULL, FALSE, arg_chunk, NULL, &error))
     {
-      g_dbus_method_invocation_return_error (invocation,
-                                             UDISKS_ERROR,
-                                             UDISKS_ERROR_FAILED,
-                                             "Error creating RAID array: %s",
-                                             error_message);
-      g_free (error_message);
+      g_prefix_error (&error, "Error creating RAID array:");
+      g_dbus_method_invocation_take_error (invocation, error);
+      udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), FALSE, error->message);
+      success = FALSE;
       goto out;
     }
 
   /* User specified name of the array, we need to get the md node */
   if (strlen (arg_name) > 0)
     {
-      /* FIXME replace by 'bd_md_node_from_name' after rewriting this to use libblockdev */
-      raid_device_name = md_node_from_name (array_name, &error);
-      if (!raid_device_name)
+      raid_node = bd_md_node_from_name (array_name, &error);
+      if (!raid_node)
         {
-          g_prefix_error (&error,
-                          "Failed to get md node for array %s",
-                          array_name);
+          g_prefix_error (&error, "Failed to get md node for array %s", array_name);
           g_dbus_method_invocation_take_error (invocation, error);
+          success = FALSE;
           goto out;
         }
-      raid_device_file = g_strdup_printf ("/dev/%s", raid_device_name);
+      raid_device_file = g_strdup_printf ("/dev/%s", raid_node);
     }
 
   else
@@ -835,6 +758,7 @@ handle_mdraid_create (UDisksManager         *_object,
                       "Error waiting for array object after creating %s",
                       raid_device_file);
       g_dbus_method_invocation_take_error (invocation, error);
+      success = FALSE;
       goto out;
     }
 
@@ -845,6 +769,7 @@ handle_mdraid_create (UDisksManager         *_object,
                                              UDISKS_ERROR_FAILED,
                                              "Error calling stat(2) on %s: %m",
                                              raid_device_file);
+      success = FALSE;
       goto out;
     }
   if (!S_ISBLK (statbuf.st_mode))
@@ -854,6 +779,7 @@ handle_mdraid_create (UDisksManager         *_object,
                                              UDISKS_ERROR_FAILED,
                                              "Device file %s is not a block device",
                                              raid_device_file);
+      success = FALSE;
       goto out;
     }
   raid_device_num = statbuf.st_rdev;
@@ -864,25 +790,17 @@ handle_mdraid_create (UDisksManager         *_object,
                            caller_uid);
 
   /* ... wipe the created RAID array */
-  if (!udisks_daemon_launch_spawned_job_sync (manager->daemon,
-                                              array_object,
-                                              "format-erase", caller_uid,
-                                              NULL, /* cancellable */
-                                              0,    /* uid_t run_as_uid */
-                                              0,    /* uid_t run_as_euid */
-                                              &status,
-                                              &error_message,
-                                              NULL, /* input_string */
-                                              "wipefs -a %s",
-                                              raid_device_file))
+  if (!bd_fs_wipe (raid_device_file, TRUE, &error))
     {
-      g_dbus_method_invocation_return_error (invocation,
-                                             UDISKS_ERROR,
-                                             UDISKS_ERROR_FAILED,
-                                             "Error wiping raid device %s: %s",
-                                             raid_device_file,
-                                             error_message);
-      goto out;
+      if (g_error_matches (error, BD_FS_ERROR, BD_FS_ERROR_NOFS))
+        g_clear_error (&error);
+      else
+        {
+          g_prefix_error (&error, "Error wiping raid device %s:", raid_device_file);
+          g_dbus_method_invocation_take_error (invocation, error);
+          success = FALSE;
+          goto out;
+        }
     }
 
   /* ... finally trigger uevents on the members - we want this so the
@@ -899,6 +817,7 @@ handle_mdraid_create (UDisksManager         *_object,
         {
           g_dbus_method_invocation_return_gerror (invocation, error);
           g_clear_error (&error);
+          success = FALSE;
           goto out;
         }
       udisks_linux_block_object_trigger_uevent (UDISKS_LINUX_BLOCK_OBJECT (object_for_block));
@@ -910,12 +829,18 @@ handle_mdraid_create (UDisksManager         *_object,
                                          invocation,
                                          g_dbus_object_get_object_path (G_DBUS_OBJECT (array_object)));
 
+  success = TRUE;
+
  out:
+
+  if (job != NULL)
+    {
+      udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), success, NULL);
+    }
+
   g_free (raid_device_file);
-  g_free (raid_device_name);
+  g_free (raid_node);
   g_free (array_name);
-  if (str != NULL)
-    g_string_free (str, TRUE);
   g_list_free_full (blocks, g_object_unref);
   g_clear_object (&array_object);
 
@@ -959,6 +884,276 @@ handle_enable_modules (UDisksManager *object,
   return TRUE; /* returning TRUE means that we handled the method invocation */
 }
 
+static gboolean
+handle_can_format (UDisksManager         *object,
+                   GDBusMethodInvocation *invocation,
+                   const gchar           *type)
+{
+  gchar *required_utility = NULL;
+  gchar *binary_path = NULL;
+
+  if (g_strcmp0 (type, "swap") == 0)
+    required_utility = g_strdup ("mkswap");
+  else if (g_strcmp0 (type, "empty") == 0)
+    required_utility = g_strdup ("wipefs");
+  else
+    {
+      const gchar **supported_fs = get_supported_filesystems ();
+      for (gsize i = 0; supported_fs[i] != NULL; i++)
+        {
+          if (g_strcmp0 (type, supported_fs[i]) == 0)
+            {
+            required_utility = g_strconcat ("mkfs.", type, NULL);
+            break;
+            }
+        }
+    }
+
+  if (required_utility == NULL)
+    {
+      g_dbus_method_invocation_return_error (invocation,
+                                             UDISKS_ERROR,
+                                             UDISKS_ERROR_NOT_SUPPORTED,
+                                             "Creation of filesystem type %s is not supported",
+                                             type);
+      return TRUE;
+    }
+
+  binary_path = g_find_program_in_path (required_utility);
+  udisks_manager_complete_can_format (object,
+                                      invocation,
+                                      g_variant_new ("(bs)", binary_path != NULL,
+                                                     binary_path != NULL ? "" : required_utility));
+  g_free (binary_path);
+  g_free (required_utility);
+
+  return TRUE;
+}
+
+static gboolean
+handle_can_resize (UDisksManager         *object,
+                   GDBusMethodInvocation *invocation,
+                   const gchar           *type)
+{
+  GError *error = NULL;
+  gchar *required_utility = NULL;
+  BDFsResizeFlags mode;
+  gboolean ret;
+
+  ret = bd_fs_can_resize (type, &mode, &required_utility, &error);
+
+  if (error != NULL)
+    {
+      g_dbus_method_invocation_take_error (invocation, error);
+      return TRUE;
+    }
+
+  udisks_manager_complete_can_resize (object,
+                                      invocation,
+                                      g_variant_new ("(bts)", ret, (guint64) mode,
+                                                     ret ? "" : required_utility));
+
+  g_free (required_utility);
+
+  return TRUE;
+}
+
+static gboolean
+handle_can_check (UDisksManager         *object,
+                  GDBusMethodInvocation *invocation,
+                  const gchar           *type)
+{
+  GError *error = NULL;
+  gchar *required_utility = NULL;
+  gboolean ret;
+
+  ret = bd_fs_can_check (type, &required_utility, &error);
+
+  if (error != NULL)
+    {
+      g_dbus_method_invocation_take_error (invocation, error);
+      return TRUE;
+    }
+
+  udisks_manager_complete_can_check (object,
+                                     invocation,
+                                     g_variant_new ("(bs)", ret,
+                                                    ret ? "" : required_utility));
+
+  g_free (required_utility);
+
+  return TRUE;
+}
+
+static gboolean
+handle_can_repair (UDisksManager         *object,
+                   GDBusMethodInvocation *invocation,
+                   const gchar           *type)
+{
+  GError *error = NULL;
+  gchar *required_utility = NULL;
+  gboolean ret;
+
+  ret = bd_fs_can_repair (type, &required_utility, &error);
+
+  if (error != NULL)
+    {
+      g_dbus_method_invocation_take_error (invocation, error);
+      return TRUE;
+    }
+
+  udisks_manager_complete_can_repair (object,
+                                      invocation,
+                                      g_variant_new ("(bs)", ret,
+                                                     ret ? "" : required_utility));
+
+  g_free (required_utility);
+
+  return TRUE;
+}
+
+
+static GSList*
+get_block_objects (UDisksManager *manager,
+                   guint         *num_blocks)
+{
+  UDisksLinuxManager *linux_manager = UDISKS_LINUX_MANAGER (manager);
+  GDBusObjectManagerServer *object_manager = NULL;
+  GList *objects = NULL;
+  GList *objects_p = NULL;
+  GSList *ret = NULL;
+
+  object_manager = udisks_daemon_get_object_manager (linux_manager->daemon);
+  objects = g_dbus_object_manager_get_objects (G_DBUS_OBJECT_MANAGER (object_manager));
+
+  for (objects_p = objects; objects_p != NULL; objects_p = objects_p->next)
+    {
+      UDisksObject *object = UDISKS_OBJECT (objects_p->data);
+      UDisksBlock *block;
+
+      block = udisks_object_get_block (object);
+      if (block != NULL)
+        {
+          ret = g_slist_prepend (ret, block);
+          (*num_blocks)++;
+        }
+    }
+
+  g_list_free_full (objects, g_object_unref);
+  ret = g_slist_reverse (ret);
+  return ret;
+}
+
+static gboolean
+handle_get_block_devices (UDisksManager         *object,
+                          GDBusMethodInvocation *invocation,
+                          GVariant              *arg_options)
+{
+  GSList *blocks = NULL;
+  GSList *blocks_p = NULL;
+  const gchar **block_paths = NULL;
+  guint num_blocks = 0;
+  guint i = 0;
+
+  blocks = get_block_objects (object, &num_blocks);
+  block_paths = g_new0 (const gchar *, num_blocks + 1);
+
+  for (i = 0,blocks_p = blocks; blocks_p != NULL; blocks_p = blocks_p->next, i++)
+      block_paths[i] = g_dbus_object_get_object_path (g_dbus_interface_get_object (G_DBUS_INTERFACE (blocks_p->data)));
+
+  udisks_manager_complete_get_block_devices  (object,
+                                              invocation,
+                                              block_paths);
+
+  g_slist_free_full (blocks, g_object_unref);
+
+  return TRUE;  /* returning TRUE means that we handled the method invocation */
+}
+
+static gboolean
+compare_paths (UDisksManager         *object,
+               UDisksBlock           *block,
+               gchar                 *path)
+{
+  const gchar *const *symlinks = NULL;
+
+  if (g_strcmp0 (udisks_block_get_device (block), path) == 0)
+    return TRUE;
+
+  symlinks = udisks_block_get_symlinks (block);
+  if (symlinks != NULL)
+    {
+      for (guint i = 0; symlinks[i] != NULL; i++)
+        if (g_strcmp0 (symlinks[i], path) == 0)
+          return TRUE;
+    }
+
+  return FALSE;
+}
+
+static gboolean
+handle_resolve_device (UDisksManager         *object,
+                       GDBusMethodInvocation *invocation,
+                       GVariant              *arg_devspec,
+                       GVariant              *arg_options)
+{
+
+  gchar *devpath = NULL;
+  gchar *devuuid = NULL;
+  gchar *devlabel = NULL;
+
+  GSList *blocks = NULL;
+  GSList *blocks_p = NULL;
+  guint num_blocks = 0;
+
+  GSList *ret = NULL;
+  GSList *ret_p = NULL;
+  guint num_found = 0;
+  const gchar **ret_paths = NULL;
+
+  gboolean found = FALSE;
+  guint i = 0;
+
+  g_variant_lookup (arg_devspec, "path", "s", &devpath);
+  g_variant_lookup (arg_devspec, "uuid", "s", &devuuid);
+  g_variant_lookup (arg_devspec, "label", "s", &devlabel);
+
+  blocks = get_block_objects (object, &num_blocks);
+
+  for (blocks_p = blocks; blocks_p != NULL; blocks_p = blocks_p->next)
+    {
+      if (devpath != NULL)
+          found = compare_paths (object, blocks_p->data, devpath);
+
+      if (devuuid != NULL)
+          found = g_strcmp0 (udisks_block_get_id_uuid (blocks_p->data), devuuid) == 0;
+      if (devlabel != NULL)
+          found = g_strcmp0 (udisks_block_get_id_label (blocks_p->data), devlabel) == 0;
+
+      if (found)
+        {
+          ret = g_slist_prepend (ret, g_object_ref (blocks_p->data));
+          num_found++;
+        }
+    }
+
+    ret_paths = g_new0 (const gchar *, num_found + 1);
+    for (i = 0,ret_p = ret; ret_p != NULL; ret_p = ret_p->next, i++)
+      {
+        ret_paths[i] = g_dbus_object_get_object_path (g_dbus_interface_get_object (G_DBUS_INTERFACE (ret_p->data)));
+      }
+
+    udisks_manager_complete_resolve_device (object,
+                                            invocation,
+                                            ret_paths);
+
+    g_slist_free_full (blocks, g_object_unref);
+    g_slist_free_full (ret, g_object_unref);
+
+    return TRUE;  /* returning TRUE means that we handled the method invocation */
+
+}
+
 /* ---------------------------------------------------------------------------------------------------- */
 
 static void
@@ -967,4 +1162,10 @@ manager_iface_init (UDisksManagerIface *iface)
   iface->handle_loop_setup = handle_loop_setup;
   iface->handle_mdraid_create = handle_mdraid_create;
   iface->handle_enable_modules = handle_enable_modules;
+  iface->handle_can_format = handle_can_format;
+  iface->handle_can_resize = handle_can_resize;
+  iface->handle_can_check = handle_can_check;
+  iface->handle_can_repair = handle_can_repair;
+  iface->handle_get_block_devices = handle_get_block_devices;
+  iface->handle_resolve_device = handle_resolve_device;
 }
