@@ -65,6 +65,7 @@ struct _UDisksClient
 typedef struct
 {
   GObjectClass parent_class;
+  GHashTable *changed_blacklist;
 } UDisksClientClass;
 
 enum
@@ -131,7 +132,7 @@ udisks_client_finalize (GObject *object)
     g_source_destroy (client->changed_timeout_source);
 
   if (client->initialization_error != NULL)
-    g_error_free (client->initialization_error);
+    g_clear_error (&(client->initialization_error));
 
   /* might be NULL if failing early in the constructor */
   if (client->object_manager != NULL)
@@ -205,6 +206,11 @@ udisks_client_class_init (UDisksClientClass *klass)
   gobject_class->finalize     = udisks_client_finalize;
   gobject_class->get_property = udisks_client_get_property;
 
+  klass->changed_blacklist = g_hash_table_new (g_str_hash, g_str_equal);
+  g_hash_table_insert (klass->changed_blacklist, (gpointer) "SmartSelftestPercentRemaining", NULL);
+  g_hash_table_insert (klass->changed_blacklist, (gpointer) "SyncRate", NULL);
+  g_hash_table_insert (klass->changed_blacklist, (gpointer) "SyncRemainingTime", NULL);
+
   /**
    * UDisksClient:object-manager:
    *
@@ -848,6 +854,66 @@ udisks_client_get_mdraid_for_block (UDisksClient  *client,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+typedef const gchar *
+(*member_retrieve_cb) (UDisksBlock *object);
+
+static GList *
+_udisks_client_get_block_or_blocks_for_mdraid (UDisksClient *client,
+                                               UDisksMDRaid *raid,
+                                               member_retrieve_cb member_get,
+                                               gboolean only_first_one,
+                                               gboolean skip_partitions)
+{
+  GList *ret = NULL;
+  GList *l, *object_proxies = NULL;
+  GDBusObject *raid_object;
+  const gchar *raid_objpath;
+
+  g_return_val_if_fail (UDISKS_IS_CLIENT (client), NULL);
+  g_return_val_if_fail (UDISKS_IS_MDRAID (raid), NULL);
+
+  raid_object = g_dbus_interface_get_object (G_DBUS_INTERFACE (raid));
+  if (raid_object == NULL)
+    goto out;
+
+  raid_objpath = g_dbus_object_get_object_path (raid_object);
+
+  object_proxies = g_dbus_object_manager_get_objects (client->object_manager);
+  for (l = object_proxies; l != NULL; l = l->next)
+    {
+      UDisksObject *object = UDISKS_OBJECT (l->data);
+      UDisksBlock *block;
+
+      block = udisks_object_get_block (object);
+      if (block == NULL)
+        continue;
+
+      /* ignore partitions */
+      if (skip_partitions)
+        {
+          if (udisks_object_peek_partition (object) != NULL)
+            continue;
+        }
+
+      if (g_strcmp0 (member_get (block), raid_objpath) == 0)
+        {
+          ret = g_list_prepend (ret, block);
+
+          if (only_first_one)
+            goto out;
+        }
+      else
+        {
+          g_object_unref (block);
+        }
+    }
+
+ out:
+  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
+  g_list_free (object_proxies);
+  return ret;
+}
+
 /**
  * udisks_client_get_block_for_mdraid:
  * @client: A #UDisksClient.
@@ -873,45 +939,19 @@ UDisksBlock *
 udisks_client_get_block_for_mdraid (UDisksClient *client,
                                     UDisksMDRaid *raid)
 {
+  GList *b_list = NULL;
   UDisksBlock *ret = NULL;
-  GList *l, *object_proxies = NULL;
-  GDBusObject *raid_object;
-  const gchar *raid_objpath;
 
-  g_return_val_if_fail (UDISKS_IS_CLIENT (client), NULL);
-  g_return_val_if_fail (UDISKS_IS_MDRAID (raid), NULL);
-
-  raid_object = g_dbus_interface_get_object (G_DBUS_INTERFACE (raid));
-  if (raid_object == NULL)
-    goto out;
-
-  raid_objpath = g_dbus_object_get_object_path (raid_object);
-
-  object_proxies = g_dbus_object_manager_get_objects (client->object_manager);
-  for (l = object_proxies; l != NULL; l = l->next)
-    {
-      UDisksObject *object = UDISKS_OBJECT (l->data);
-      UDisksBlock *block;
-
-      block = udisks_object_get_block (object);
-      if (block == NULL)
-        continue;
-
-      /* ignore partitions */
-      if (udisks_object_peek_partition (object) != NULL)
-        continue;
-
-      if (g_strcmp0 (udisks_block_get_mdraid (block), raid_objpath) == 0)
-        {
-          ret = block;
-          goto out;
-        }
-      g_object_unref (block);
-    }
-
- out:
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
+  b_list = _udisks_client_get_block_or_blocks_for_mdraid (client,
+                                                          raid,
+                                                          udisks_block_get_mdraid,
+                                                          TRUE,   /* Retrieve first one */
+                                                          TRUE);  /* Skip partitions */
+  if (b_list)
+   {
+      ret = (g_list_first (b_list))->data;
+      g_list_free (b_list);
+   }
   return ret;
 }
 
@@ -938,49 +978,11 @@ GList *
 udisks_client_get_all_blocks_for_mdraid (UDisksClient *client,
                                          UDisksMDRaid *raid)
 {
-  GList *ret = NULL;
-  GList *l, *object_proxies = NULL;
-  GDBusObject *raid_object;
-  const gchar *raid_objpath;
-
-  g_return_val_if_fail (UDISKS_IS_CLIENT (client), NULL);
-  g_return_val_if_fail (UDISKS_IS_MDRAID (raid), NULL);
-
-  raid_object = g_dbus_interface_get_object (G_DBUS_INTERFACE (raid));
-  if (raid_object == NULL)
-    goto out;
-
-  raid_objpath = g_dbus_object_get_object_path (raid_object);
-
-  object_proxies = g_dbus_object_manager_get_objects (client->object_manager);
-  for (l = object_proxies; l != NULL; l = l->next)
-    {
-      UDisksObject *object = UDISKS_OBJECT (l->data);
-      UDisksBlock *block;
-
-      block = udisks_object_get_block (object);
-      if (block == NULL)
-        continue;
-
-      /* ignore partitions */
-      if (udisks_object_peek_partition (object) != NULL)
-        continue;
-
-      if (g_strcmp0 (udisks_block_get_mdraid (block), raid_objpath) == 0)
-        {
-          ret = g_list_prepend (ret, block);
-        }
-      else
-        {
-          g_object_unref (block);
-        }
-    }
-
- out:
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
-  ret = g_list_reverse (ret);
-  return ret;
+  return g_list_reverse (_udisks_client_get_block_or_blocks_for_mdraid (client,
+                                                                        raid,
+                                                                        udisks_block_get_mdraid,
+                                                                        FALSE,    /* Retrieve all */
+                                                                        TRUE));   /* Skip partitions */
 }
 
 /**
@@ -1000,46 +1002,12 @@ GList *
 udisks_client_get_members_for_mdraid (UDisksClient *client,
                                       UDisksMDRaid *raid)
 {
-  GList *ret = NULL;
-  GList *l, *object_proxies = NULL;
-  GDBusObject *raid_object;
-  const gchar *raid_objpath;
-
-  g_return_val_if_fail (UDISKS_IS_CLIENT (client), NULL);
-  g_return_val_if_fail (UDISKS_IS_MDRAID (raid), NULL);
-
-  raid_object = g_dbus_interface_get_object (G_DBUS_INTERFACE (raid));
-  if (raid_object == NULL)
-    goto out;
-
-  raid_objpath = g_dbus_object_get_object_path (raid_object);
-
-  object_proxies = g_dbus_object_manager_get_objects (client->object_manager);
-  for (l = object_proxies; l != NULL; l = l->next)
-    {
-      UDisksObject *object = UDISKS_OBJECT (l->data);
-      UDisksBlock *block;
-
-      block = udisks_object_get_block (object);
-      if (block == NULL)
-        continue;
-
-      if (g_strcmp0 (udisks_block_get_mdraid_member (block), raid_objpath) == 0)
-        {
-          ret = g_list_prepend (ret, block); /* adopts reference to block */
-        }
-      else
-        {
-          g_object_unref (block);
-        }
-    }
-
- out:
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
-  return ret;
+  return _udisks_client_get_block_or_blocks_for_mdraid (client,
+                                                        raid,
+                                                        udisks_block_get_mdraid_member,
+                                                        FALSE,    /* Retrieve all */
+                                                        FALSE);   /* Don't skip partitions */
 }
-
 /* ---------------------------------------------------------------------------------------------------- */
 
 static void
@@ -1589,7 +1557,25 @@ on_interface_proxy_properties_changed (GDBusObjectManagerClient   *manager,
                                        gpointer                    user_data)
 {
   UDisksClient *client = UDISKS_CLIENT (user_data);
-  udisks_client_queue_changed (client);
+  UDisksClientClass *client_class = UDISKS_CLIENT_GET_CLASS (client);
+
+  GVariantIter iter;
+  gchar *property_name = NULL;
+
+  /* never emit the change signal for Job objects */
+  if (g_strcmp0 (g_dbus_proxy_get_interface_name (interface_proxy), "org.freedesktop.UDisks2.Drive.Job") == 0)
+    return;
+
+  g_variant_iter_init (&iter, changed_properties);
+  while (g_variant_iter_next (&iter, "{&sv}", &property_name, NULL))
+    {
+      if (! g_hash_table_contains (client_class->changed_blacklist, property_name))
+        {
+          /* one of the properties is not on the blacklist -> emit change signal */
+          udisks_client_queue_changed (client);
+          return;
+        }
+    }
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -1979,9 +1965,13 @@ udisks_client_get_id_for_display (UDisksClient *client,
                                   gboolean      long_string)
 {
   guint n;
-  gchar *ret;
+  gchar *ret = NULL;
 
-  ret = NULL;
+  if (usage == NULL || type == NULL || version == NULL)
+    {
+      g_critical ("Invalid device ID specification. Usage, type and version must not be NULL");
+      goto out;
+    }
 
   for (n = 0; id_type[n].usage != NULL; n++)
     {
@@ -2325,7 +2315,7 @@ static const struct
   {"dos", "microsoft", "0x01",  NC_("part-type", "FAT12"), 0},
   {"dos", "microsoft", "0x04",  NC_("part-type", "FAT16 <32M"), 0},
   {"dos", "microsoft", "0x06",  NC_("part-type", "FAT16"), 0},
-  {"dos", "microsoft", "0x07",  NC_("part-type", "HPFS/NTFS"), 0},
+  {"dos", "microsoft", "0x07",  NC_("part-type", "NTFS/exFAT/HPFS"), 0},
   {"dos", "microsoft", "0x0b",  NC_("part-type", "W95 FAT32"), 0},
   {"dos", "microsoft", "0x0c",  NC_("part-type", "W95 FAT32 (LBA)"), 0},
   {"dos", "microsoft", "0x0e",  NC_("part-type", "W95 FAT16 (LBA)"), 0},
@@ -2503,9 +2493,13 @@ udisks_client_get_job_description (UDisksClient   *client,
       g_hash_table_insert (hash, (gpointer) "encrypted-modify",     (gpointer) C_("job", "Modifying Encrypted Device"));
       g_hash_table_insert (hash, (gpointer) "swapspace-start",      (gpointer) C_("job", "Starting Swap Device"));
       g_hash_table_insert (hash, (gpointer) "swapspace-stop",       (gpointer) C_("job", "Stopping Swap Device"));
+      g_hash_table_insert (hash, (gpointer) "swapspace-modify",     (gpointer) C_("job", "Modifying Swap Device"));
+      g_hash_table_insert (hash, (gpointer) "filesystem-check",     (gpointer) C_("job", "Checking Filesystem"));
       g_hash_table_insert (hash, (gpointer) "filesystem-mount",     (gpointer) C_("job", "Mounting Filesystem"));
       g_hash_table_insert (hash, (gpointer) "filesystem-unmount",   (gpointer) C_("job", "Unmounting Filesystem"));
       g_hash_table_insert (hash, (gpointer) "filesystem-modify",    (gpointer) C_("job", "Modifying Filesystem"));
+      g_hash_table_insert (hash, (gpointer) "filesystem-repair",    (gpointer) C_("job", "Repairing Filesystem"));
+      g_hash_table_insert (hash, (gpointer) "filesystem-resize",    (gpointer) C_("job", "Resizing Filesystem"));
       g_hash_table_insert (hash, (gpointer) "format-erase",         (gpointer) C_("job", "Erasing Device"));
       g_hash_table_insert (hash, (gpointer) "format-mkfs",          (gpointer) C_("job", "Creating Filesystem"));
       g_hash_table_insert (hash, (gpointer) "loop-setup",           (gpointer) C_("job", "Setting Up Loop Device"));
